/*
 * Copyright (C) 2011 Giorgio Vazzana
 *
 * This file is part of hwebcam.
 *
 * hwebcam is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <sys/time.h>
#include <theora/theoraenc.h>
#include "video.h"

#ifdef USE_SDL
#include "SDL.h"
#endif

#define VERSION "0.2"

typedef struct ppm_st {
	unsigned int   width;
	unsigned int   height;
	unsigned int   max;
	unsigned char  *data;
} ppm;

int keep_going = 2;

void signal_handler(int sig)
{
	(void) sig;

	keep_going = 1;
}

#define CLIPRGB(x) if(x < 0) x = 0; else if(x > 255) x = 255;
#define COMPUTE_RGB(y, cb, cr) r = (256 * y +          + 359 * cr + 128) >> 8;\
                               g = (256 * y -  88 * cb - 183 * cr + 128) >> 8;\
                               b = (256 * y + 454 * cb +          + 128) >> 8;
void yuyv422_to_rgb24(const unsigned char *src, ppm *frame)
{
	unsigned int i, j;
	int y1, cb, y2, cr;
	int r, g, b;

	for (i=0; i<frame->height; i++) {
		for (j=0; j<frame->width; j+=2) {
			y1 = *(src++);
			cb = *(src++) - 128;
			y2 = *(src++);
			cr = *(src++) - 128;

			COMPUTE_RGB(y1, cb, cr)
			CLIPRGB(r)
			CLIPRGB(g)
			CLIPRGB(b)

			*(frame->data + 3*j + 3*i*frame->width + 0) = r;
			*(frame->data + 3*j + 3*i*frame->width + 1) = g;
			*(frame->data + 3*j + 3*i*frame->width + 2) = b;

			COMPUTE_RGB(y2, cb, cr)
			CLIPRGB(r)
			CLIPRGB(g)
			CLIPRGB(b)

			*(frame->data + 3*(j+1) + 3*i*frame->width + 0) = r;
			*(frame->data + 3*(j+1) + 3*i*frame->width + 1) = g;
			*(frame->data + 3*(j+1) + 3*i*frame->width + 2) = b;
		}
	}
}

void yuyv422_to_yuv420p(const unsigned char *src, th_ycbcr_buffer ycbcrbuf)
{
	int i, j, y1, cb, y2, cr;

	for (i=0; i<ycbcrbuf[0].height; i++) {
		for (j=0; j<ycbcrbuf[0].width; j+=2) {
			y1 = *(src++);
			cb = *(src++);
			y2 = *(src++);
			cr = *(src++);

			*(ycbcrbuf[0].data + i*ycbcrbuf[0].width + j  ) = y1;
			*(ycbcrbuf[0].data + i*ycbcrbuf[0].width + j+1) = y2;

			if (!(i&1)) {
				cb += *(src + 2 * ycbcrbuf[0].width - 4 + 1);
				*(ycbcrbuf[1].data + (i>>1)*ycbcrbuf[1].width + (j>>1)) = cb >> 1;

				cr += *(src + 2 * ycbcrbuf[0].width - 4 + 3);
				*(ycbcrbuf[2].data + (i>>1)*ycbcrbuf[2].width + (j>>1)) = cr >> 1;
			}
		}
	}
}

void usage(void)
{
#ifdef USE_SDL
	fprintf(stderr, "usage: hwebcam [-h -v -d -s -i -t -e -o -b -q -w]\n");
#else
	fprintf(stderr, "usage: hwebcam [-h -v -d -s -i -t -e -o -b -q]\n");
#endif
	fprintf(stderr, "       -h          print help\n");
	fprintf(stderr, "       -v          verbose\n");
	fprintf(stderr, "       -d device   use video device  (/dev/video0)\n");
	fprintf(stderr, "       -s WxH      choose frame size (640x400)\n");
	fprintf(stderr, "       -i fps      choose frame rate (20)\n");
	fprintf(stderr, "       -t val      save a frame every val seconds (disabled, 0=every frame)\n");
	fprintf(stderr, "       -e          encode captured stream in ogg/theora\n");
	fprintf(stderr, "       -o ofile    save captured stream in ofile (capture.ogg, -=stdout)\n");
	fprintf(stderr, "       -b bitrate  choose theora bitrate in bit/s (200000)\n");
	fprintf(stderr, "       -q quality  choose theora quality (0, accepted values [0-63])\n");
#ifdef USE_SDL
	fprintf(stderr, "       -w          display stream in a sdl window\n");
#endif
	exit(1);
}

int main(int argc, char *argv[])
{
	int ret, opt, i, j;
	int verbose = 0, width = 640, height = 400, fps = 20, timeout = -1;
	char *device = "/dev/video0";
	struct video_data vd;
	struct sigaction act;
	struct timeval frame_start, frame_stop;
	time_t timeout_start, timeout_stop;
	long frame_interval;
	double actual_fps;
	ppm ppmimage;
	FILE *fpimage = NULL;
	char imagename[32] = "image.ppm";

	/* ogg data */
	ogg_stream_state oss;
	ogg_packet opacket;
	ogg_page opage;
	/* theora data */
	th_info ti;
	th_comment tc;
	th_enc_ctx *tctx = NULL;
	th_ycbcr_buffer ycbcrbuf;
	int encode = 0, bitrate = 200000, quality = 0, duplicate = 0;
	FILE *fpvideo = NULL;
	char *videoname = "capture.ogg";

#ifdef USE_SDL
	int sdl_window = 0;
	char caption_string[32];
	SDL_Surface* surface = NULL;
	SDL_Overlay *overlay = NULL;
	SDL_Rect     video_rect;
#endif

	fprintf(stderr, "hwebcam, ver. %s\n\n", VERSION);

#ifdef USE_SDL
	while ((opt = getopt(argc, argv, "hvd:s:i:t:eo:b:q:w")) != -1) {
#else
	while ((opt = getopt(argc, argv, "hvd:s:i:t:eo:b:q:")) != -1) {
#endif
		switch (opt) {
		case 'h':
			usage();
			break;
		case 'v':
			verbose = 1;
			break;
		case 'd':
			device = optarg;
			break;
		case 's':
			ret = sscanf(optarg, "%dx%d", &width, &height);
			if (ret != 2) {
				fprintf(stderr, "Error: cannot read frame size\n\n");
				usage();
			}
			break;
		case 'i':
			fps = atoi(optarg);
			break;
		case 't':
			timeout = atoi(optarg);
			break;
		case 'e':
			encode = 1;
			break;
		case 'o':
			videoname = optarg;
			break;
		case 'b':
			bitrate = atoi(optarg);
			break;
		case 'q':
			quality = atoi(optarg);
			bitrate = 0;
			break;
#ifdef USE_SDL
		case 'w':
			sdl_window = 1;
			break;
#endif
		default:
			fprintf(stderr, "\n");
			usage();
		}
	}

/*	fprintf(stderr, "v=%d, d=%s, s=%dx%d, i=%d, t=%d, e=%d, o=%s, b=%d, q=%d, w=%d\n",
			verbose, device, width, height, fps, timeout, encode, videoname, bitrate, quality, sdl_window); */

	if (width<0 || width>1280 || height<0 || height>1024 || fps<0 || fps>30 || quality<0 || quality>63) {
		fprintf(stderr, "Error: parameters out of bounds\n\n");
		usage();
	}

	/* Install signal handler */
	act.sa_handler = signal_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags = SA_RESTART;
	sigaction(SIGINT, &act, NULL);

	/* Init video */
	ret = video_init(&vd, device, width, height, fps, verbose);
	if (ret == -1)
		exit(1);
	if ((unsigned int)width != vd.width || (unsigned int)height != vd.height || (unsigned int)fps != vd.tpfden) {
		fprintf(stderr, "Warning: device do not support %dx%d@%d, %dx%d@%d chosen instead\n",
				width, height, fps, vd.width, vd.height, vd.tpfden);
		width  = vd.width;
		height = vd.height;
		fps    = vd.tpfden;
	}

	/* Prepare ppm image */
	if (timeout >= 0) {
		ppmimage.width  = width;
		ppmimage.height = height;
		ppmimage.max    = 255;
		ppmimage.data   = malloc(3 * ppmimage.width * ppmimage.height * sizeof(unsigned char));
		if (ppmimage.data == NULL)  {
			fprintf(stderr, "Error: malloc() failed\n");
			exit(1);
		}
	}

	/* Prepare theora encoder */
	if (encode) {
		/* open output file */
		if (!strcmp(videoname, "-"))
			fpvideo = stdout;
		else {
			fpvideo = fopen(videoname, "w");
			if (fpvideo == NULL) {
				fprintf(stderr, "Error: cannot open %s\n", videoname);
				exit(1);
			}
		}

		/* prepare ogg stream, serialno = time(NULL) */
		ogg_stream_init(&oss, time(NULL));

		/* configure theora encoder */
		th_info_init(&ti);
		th_comment_init(&tc);
		th_comment_add(&tc, "Comment=hwebcam stream");

		ti.frame_width            = width;
		ti.frame_height           = height;
		ti.pic_width              = width;
		ti.pic_height             = height;
		ti.pic_x                  = 0;
		ti.pic_y                  = 0;
		ti.colorspace             = TH_CS_UNSPECIFIED;
		ti.pixel_fmt              = TH_PF_420;
		ti.target_bitrate         = bitrate;
		ti.quality                = quality;
		ti.keyframe_granule_shift = 6;
		ti.version_major          = 0;
		ti.version_minor          = 0;
		ti.version_subminor       = 0;
		ti.fps_numerator          = fps;
		ti.fps_denominator        = 1;
		ti.aspect_numerator       = 1;
		ti.aspect_denominator     = 1;

		/* allocate encoder handle */
		tctx = th_encode_alloc(&ti);
		if (tctx == NULL) {
			fprintf(stderr, "Error: th_encode_alloc() failed\n");
			exit(1);
		}

		/* header packets */
		while ((ret = th_encode_flushheader(tctx, &tc, &opacket)) > 0) {
			if (ogg_stream_packetin(&oss, &opacket) == 0) {
				//if (verbose) fprintf(stderr, "   H\tP");
				while (ogg_stream_pageout(&oss, &opage)) {		/* pageout or flush? */
					//if (verbose) fprintf(stderr, "p");
					fwrite(opage.header, (size_t) 1, (size_t) opage.header_len, fpvideo);
					fwrite(opage.body, (size_t) 1, (size_t) opage.body_len, fpvideo);
				}
				//if (verbose) fprintf(stderr, "\n");
			} else {
				fprintf(stderr, "Error: ogg_stream_packein() failed\n");
				exit(1);
			}
		}
		if (ret == TH_EFAULT) {
			fprintf(stderr, "Error: th_encode_flushheader() failed\n");
			exit(1);
		}
		/* force remaining header packets into a page */
		while (ogg_stream_flush(&oss, &opage)) {
			//if (verbose) fprintf(stderr, "pf");
			fwrite(opage.header, (size_t) 1, (size_t) opage.header_len, fpvideo);
			fwrite(opage.body, (size_t) 1, (size_t) opage.body_len, fpvideo);
		}

		/* set up theora's ycbcr buffer */
		ycbcrbuf[0].width  = width;
		ycbcrbuf[0].height = height;
		ycbcrbuf[0].stride = ycbcrbuf[0].width;
		ycbcrbuf[0].data   = malloc(ycbcrbuf[0].width * ycbcrbuf[0].stride * sizeof(unsigned char));

		ycbcrbuf[1].width  = width/2;
		ycbcrbuf[1].height = height/2;
		ycbcrbuf[1].stride = ycbcrbuf[1].width;
		ycbcrbuf[1].data   = malloc(ycbcrbuf[1].width * ycbcrbuf[1].stride * sizeof(unsigned char));

		ycbcrbuf[2].width  = width/2;
		ycbcrbuf[2].height = height/2;
		ycbcrbuf[2].stride = ycbcrbuf[2].width;
		ycbcrbuf[2].data   = malloc(ycbcrbuf[2].width * ycbcrbuf[2].stride * sizeof(unsigned char));

		if (!ycbcrbuf[0].data || !ycbcrbuf[1].data || !ycbcrbuf[2].data) {
			fprintf(stderr, "Error: malloc() failed\n");
			exit(1);
		}
	}

#ifdef USE_SDL
	if (sdl_window) {
		/* Initialize SDL Video subsystem */
		if (verbose)
			fprintf(stderr, "Initializing SDL...\n");
		if(SDL_Init(SDL_INIT_VIDEO) == -1) {
			fprintf(stderr, "Error: could not initialize SDL: %s\n", SDL_GetError());
			exit(1);
		}

		/* Clean up on exit */
		atexit(SDL_Quit);

		/* Initialize the display requesting a software surface */
		surface = SDL_SetVideoMode(width, height, 24, SDL_SWSURFACE|SDL_ANYFORMAT);
		if (surface == NULL) {
			fprintf(stderr, "Error: couldn't set requested video mode: %s\n", SDL_GetError());
			exit(1);
		}
		if (verbose)
			fprintf(stderr, "  Surface: %dx%d at %d bits-per-pixel mode\n", surface->w, surface->h, surface->format->BitsPerPixel);

		/* Create a YUV video overlay  */
		overlay = SDL_CreateYUVOverlay(width, height, SDL_YUY2_OVERLAY, surface);
		if (verbose) {
			fprintf(stderr, "  Overlay:\n");
			fprintf(stderr, "    Dimensions:     %dx%d\n", overlay->w, overlay->h);
			fprintf(stderr, "    Format:         0x%08x\n", overlay->format);
			fprintf(stderr, "    Planes:         %d\n", overlay->planes);
			for (i=0; i<overlay->planes; i++)
				fprintf(stderr, "    Pitches[%d]:     %d\n", i, overlay->pitches[i]);
			fprintf(stderr, "    Hw accelerated: %d\n", overlay->hw_overlay);
		}

		/* Set the window title and icon name */
		SDL_WM_SetCaption("hwebcam", "hwebcam");
	}
#endif

	/* Print streaming info */
	fprintf(stderr, "Info:\n");
	fprintf(stderr, "  Device:     %s\n", device);
	fprintf(stderr, "  Frame size: %dx%d\n", width, height);
	fprintf(stderr, "  Frame rate: %d fps\n", fps);
	if (timeout >= 0)
	fprintf(stderr, "  Save frame: every %d seconds\n", timeout);
#ifdef USE_SDL
	fprintf(stderr, "  Window:     %s\n", sdl_window ? "SDL" : "none");
#endif
	if (encode) {
	fprintf(stderr, "Theora stream:\n");
	fprintf(stderr, "  libtheora:  %s\n", th_version_string());
	fprintf(stderr, "  File name:  %s\n", strcmp(videoname, "-") == 0 ? "stdout" : videoname);
	fprintf(stderr, "  Bitrate:    %d bit/s\n", ti.target_bitrate);
	fprintf(stderr, "  Quality:    %d\n", ti.quality);
	}

	/* Main loop */
	frame_start.tv_sec  = 0;	//silence warning
	frame_start.tv_usec = 0;
	timeout_start = time(NULL);
	for (i = 0; keep_going; i++) {
		/* get a video frame */
		ret = video_frame_get(&vd);
		if (ret == -1)
			exit(1);

		/* compute frame time and actual fps */
		frame_stop     = vd.buffer.timestamp;
		if (i == 0)
			frame_start= frame_stop;
		frame_interval = (frame_stop.tv_sec - frame_start.tv_sec) * 1000000L + frame_stop.tv_usec - frame_start.tv_usec;
		actual_fps     = (frame_interval == 0) ? fps : (1000000.0 / (double)frame_interval);
		/* print stats in verbose mode every frame if actual_fps<=10, or every 2 frames */
		if (verbose && ((int)(actual_fps+0.5)<=10 || !(i%2))) {
			fprintf(stderr, "\r  Frame: %5d, Buffer: %2d, Interval: %7ld us, Fps: %5.2f", i, vd.buffer.index, frame_interval, actual_fps);
			fflush(stdout);
		}
		frame_start    = vd.buffer.timestamp;

#ifdef USE_SDL
		if (sdl_window) {
			/* Fill in pixel data */
			SDL_LockYUVOverlay(overlay);
			for(j=0; j<height; j++)
				memcpy(overlay->pixels[0]+j*overlay->pitches[0], vd.buffers[vd.buffer.index].start+j*overlay->pitches[0], (size_t) overlay->pitches[0]);
			SDL_UnlockYUVOverlay(overlay);
			/* Display frame */
			video_rect.x = 0;
			video_rect.y = 0;
			video_rect.w = width;
			video_rect.h = height;
			SDL_DisplayYUVOverlay(overlay, &video_rect);
			/* Print fps in window's title, every 1 second */
			if ((i % (int)(actual_fps+0.5)) == 0) {
				sprintf(caption_string, "hwebcam - %5.2f fps", actual_fps);
				SDL_WM_SetCaption(caption_string, "hwebcam");
			}
		}
#endif

		if (encode) {
			/* prepare the frame for the encoder */
			yuyv422_to_yuv420p(vd.buffers[vd.buffer.index].start, ycbcrbuf);

			/* if actual_fps==fps/2, push 2 frames into the encoder. if fps/4, push 4 frames */
			duplicate = 0;
			if (0.45 * (double)fps < actual_fps && actual_fps < 0.55 * (double)fps)
				duplicate = 1;
			else if (0.20 * (double)fps < actual_fps && actual_fps < 0.30 * (double)fps)
				duplicate = 3;

			/* set the number of duplicates of the next frame to produce */
			if (duplicate)
				if (th_encode_ctl(tctx, TH_ENCCTL_SET_DUP_COUNT, &duplicate, sizeof(duplicate)) != 0) {
					fprintf(stderr, "Error: th_encode_ctl() failed\n");
					exit(1);
				}

			/* pass the frame to the encoder */
			if (th_encode_ycbcr_in(tctx, ycbcrbuf) != 0) {
				fprintf(stderr, "Error: th_encode_ycbcr_in() failed\n");
				exit(1);
			}

			/* get the packet out and pass it to the ogg stream */
			while ((ret = th_encode_packetout(tctx, keep_going==1 ? (keep_going--) : 0, &opacket)) > 0) {
				if (ogg_stream_packetin(&oss, &opacket) == 0) {
					//if (verbose) fprintf(stderr, "P");
					while (ogg_stream_pageout(&oss, &opage)) {
						//if (verbose) fprintf(stderr, "p");
						fwrite(opage.header, (size_t) 1, (size_t) opage.header_len, fpvideo);
						fwrite(opage.body, (size_t) 1, (size_t) opage.body_len, fpvideo);
					}
				} else {
					fprintf(stderr, "Error: ogg_stream_packein() failed\n");
					exit(1);
				}
			}
			if (ret == TH_EFAULT) {
				fprintf(stderr, "Error: th_encode_packetout() failed\n");
				exit(1);
			}
		}

		if (timeout >= 0) {
			timeout_stop = time(NULL);
			if (difftime(timeout_stop, timeout_start) >= timeout) {
				/* Dump data to ppm image */
				yuyv422_to_rgb24(vd.buffers[vd.buffer.index].start, &ppmimage);
				//sprintf(imagename, "%08d.ppm", i);
				fpimage = fopen(imagename, "w");
				if (fpimage == NULL)  {
					fprintf(stderr, "Error: cannot open %s\n", imagename);
					exit(1);
				}
				fprintf(fpimage, "P6\n%u %u\n%u\n", ppmimage.width, ppmimage.height, ppmimage.max);
				fwrite(ppmimage.data, 3, ppmimage.width * ppmimage.height, fpimage);
				fclose(fpimage);

				if (verbose)
					fprintf(stderr, "  Image saved\n");
				timeout_start = time(NULL);
			}
		}

		/* ok, we're done with this frame */
		ret = video_frame_done(&vd);
		if (ret == -1)
			exit(1);

		if (!encode && keep_going == 1)
			keep_going = 0;
	}
	fprintf(stderr, "\n");

	ret = video_close(&vd);

	if (encode) {
		/* flush last packets, if any */
		while (ogg_stream_flush(&oss, &opage)) {
			//if (verbose) fprintf(stderr, "pf");
			fwrite(opage.header, (size_t) 1, (size_t) opage.header_len, fpvideo);
			fwrite(opage.body, (size_t) 1, (size_t) opage.body_len, fpvideo);
		}
		th_encode_free(tctx);
		ogg_stream_clear(&oss);
		if (strcmp(videoname, "-"))
			fclose(fpvideo);
	}

	return 0;
}
